/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Source for class which maps from integer indices to elements
// of class T. 
  
#if ! defined _map_h_
#define _map_h_ 1

// Define the macro below (at your peril!) to obtain reams of debug
// output from the BaseMap class.

// #define DEBUG_MAPS

#include "Lists.h"

#define DEFAULT_MAP_SIZE 1   // initial size of a map when created
#define EXPANSION	 3   // expansion coeffiecient when map grows
#define REDUCTION	 2   // contraction coefficient when map shrinks

// CLASS GROUP: the Map container classes

template <class K, class T> class Base_Map_Iterator ;
template <class K, class T> class ObjectMap_Iterator ;

/*--------------------------------------------------------------------+

CLASS
    BaseMap

    Implements mappings from class K objects to class T objects

DESCRIPTION

 The Map and NameMap classes use this class to construct mappings, from
 integers and strings respectively, to some given class T. 
 The BaseMap class is implemented using dynamically sized hash tables
 which grow and shrink to ensure that the table size is always
 within a constant fraction of the table loading. This ensures O(1)
 insertion, deletion and retrieval times.

 The hash tables are constructed using the class slist&ltT&gt
 for simplicity and clarity.

 Class T must provide an integer or char* function called index()
 which returns a unique identifying key for each object
 of that class that may exist within the map.

 Each map records the following values:

  (a). nitems  - the number of stored items in the map
 
  (b). th_high - when nitems exceeds this value the map expands

  (c). th_low  - when nitems less than this, the map contracts

 The dynamic sizing always ensures that the number of hashing bins
 satisfies the following inequalities:

         nitems * REDUCTION &lt bins &lt nitems / EXPANSION

 Thus, to guarantee a loading on the hash bins of 0.5 to 3.0, one
 can set REDUCTION=2 and EXPANSION=3 (and this is the default).

 You must ensure that REDUCTION &lt= EXPANSION, and if REDUCTION is
 strictly &lt EXPANSION then there will be hysteresis in the way
 expansions and contractions occur. This prevents pathalogical
 behaviour if nitems alternates around th_high and th_high-1.

--------------------------------------------------------------------*/

template <class K, class T> class BaseMap {

friend class Base_Map_Iterator<K, T> ;

public:
  
	/////////////////////////////////
	//
	// Returns count of items in the map
	//
  int contents () { return items ; }

	/////////////////////////////////
	//
	// Empties the map, but does not destroy the items
	//
  void clear () {
    if (htab)
      for (int i = 0; i < msize; i++)
	htab[i].clear() ;
  }

	/////////////////////////////////
	//
	// Empties the map AND destroys the contained items
	//
  void zap () {
#ifdef DEBUG_MAPS
    printf (" Map-zap:\n") ;
#endif
    if (htab)
      for (int i = 0; i < msize; i++)
	{
#ifdef DEBUG_MAPS
	  printf ("  bin-zap [%d]\n", i) ;
#endif
	  htab[i].zap() ;
	}
  }

	/////////////////////////////////
	//
	// Adds an element to the map - passed by reference
	//
	//  [in] elem   the class T element to be added
	//
  void operator+= (T& elem) {
    htab[hash(elem.index())].prepend(&elem) ;
    if (++items > th_high)
      resize (msize * EXPANSION + 1) ;
  }

	/////////////////////////////////
	//
	// Adds an element to the map - passed by pointer
	//
	//  [in] elem   pointer to the class T element to be added
	//
  void operator+= (T* elem) {
    if (elem) {
#ifdef DEBUG_MAPS
      printf ("...Map += item 0x%08x\n", (unsigned long)elem) ;
#endif
      htab[hash(elem->index())].prepend(elem) ;
      if (++items > th_high)
	resize (msize * EXPANSION + 1) ;
    }
  }

	/////////////////////////////////
	//
        // Search for element identified by 'key' and
        // return a pointer to it if it exists;
	// otherwise return zero.
	//
	//  [in] key  a class K object acting as search index
	//
  T* operator[] (K key) {
    return htab[hash(key)].find(key) ;
  }

	/////////////////////////////////
	//
        // Search for element identified by 'key' 
        // which has variant property 'v', and
        // return a pointer to it if it exists; 
        // otherwise return zero.
	// 
	//  [in] key  a class K object acting as search index
        //  [in] v    an integer which must match the variant attribute
	//
  T* lookup_variant (K key, int v) {

    return htab[hash(key)].find(key, v) ;
  }

	/////////////////////////////////
	//
        // Search for element identified by 'key' 
        // which has variant property 'v', and
        // remove it from the map if it exists.
	// 
	//  [in] key  a class K object acting as search index
        //  [in] v    an integer which must match the variant attribute
	//
  void remove_variant (K key, int v) {

    if ((items -= htab[hash(key)].remove(key, v)) < th_low)
      resize (msize / REDUCTION + 1) ;
  }

	/////////////////////////////////
	//
        // Search for element 'elem' 
        // and remove it from the map if it exists.
	// 
	//  [in] elem  the class T object to be removed
	//
  void remove (T& elem) {
    if ((items -= htab[hash(elem.index())].remove(elem.index())) < th_low)
      resize (msize / REDUCTION + 1) ;
  }

	/////////////////////////////////
	//
        // Search for element pointed to by 'elem' 
        // and remove it from the map if it exists.
	// 
	//  [in] elem  pointer to the class T object to be removed
	//
  void remove (T* elem) {
    if (elem)
      if ((items -= htab[hash(elem->index())].remove(elem->index())) < th_low)
	resize (msize / REDUCTION + 1) ;
  }

	/////////////////////////////////
	//
	// Print out a representation of the map on stdout
	//
  void print () {
    int i ;
    for (i = 0 ; i < msize ; i++)
    {
      printf ("  htab [%d]: ", i) ;
      htab[i].print() ;
      printf ("\n") ;
    }  
  }

	/////////////////////////////////
	//
	// Resize the map's hash table.
	//
	//  [in] newsize   the number of entries in the resized table
	//
  void resize (int newsize) {
#ifdef DEBUG_MAPS
    printf ("Resizing BaseMap 0x%09x from %d to %d\n",
	    (unsigned long)this, msize, newsize) ;
#endif
    if (newsize < 1)
      return ;

#ifdef DEBUG_MAPS
    printf (" Original BaseMap is:-\n") ;
    print () ;
#endif
    BaseMap<K,T> newmap (*this, newsize) ;
#ifdef DEBUG_MAPS
    printf (" ... new large Map has been created from the old\n") ;
#endif
    delete [] htab ;
#ifdef DEBUG_MAPS
    printf (" ... deleted [] old htab\n") ;
#endif
    msize = newsize ;
    htab = newmap.htab ;
    newmap.htab = 0 ;
#ifdef DEBUG_MAPS
    printf (" ... copied htab from new Map\n") ;
#endif
    th_high = newmap.th_high ;
    th_low = newmap.th_low ;
#ifdef DEBUG_MAPS
    printf (" ... resizing complete.\n") ;
    printf (" Resulting BaseMap is:-\n") ;
    print () ;
#endif
  }

	/////////////////////////////////
	//
	// Default constructor
	//
	//  [in] sz optional initial table size
	//
  BaseMap (int sz = DEFAULT_MAP_SIZE) {
#ifdef DEBUG_MAPS
    printf ("BaseMap construction for 0x%09x [%d]\n", (unsigned long)this, sz) ;
#endif
    htab = new slist<K, T> [sz] ;
#ifdef DEBUG_MAPS
    printf ("  htab = (0x%09x..0x%09x)\n", (unsigned long)htab, (unsigned long)&htab[sz-1]) ;
#endif
    msize = sz ;
    items = 0 ;
    th_low = sz / REDUCTION ;
    th_high = sz * EXPANSION ;
  }

	/////////////////////////////////
	//
	// Copy constructor
	//
	//  [in] m  an existing BaseMap object
	//  [in] sz an (optional) size for the hash table
	//
  BaseMap (BaseMap<K, T>& m, int sz = DEFAULT_MAP_SIZE) {
#ifdef DEBUG_MAPS
    printf ("BaseMap copy from 0x%09x to 0x%09x [%d]\n", 
	    (unsigned long)&m, (unsigned long)this, sz) ;
#endif
    htab = new slist<K, T> [sz] ;
#ifdef DEBUG_MAPS
    printf ("  htab = (0x%09x..0x%09x)\n", (unsigned long)htab, (unsigned long)&htab[sz-1]) ;
#endif
    msize = sz ;
    items = 0 ;
    th_low = sz / REDUCTION ;
    th_high = sz * EXPANSION ;
    Base_Map_Iterator<K,T> next_item (m) ;
    T* item ;
    
    while ((item = next_item()))
      (*this) += item ;
  }

	/////////////////////////////////
	//
	// Destructor
	//
  ~BaseMap() { 
#ifdef DEBUG_MAPS
    printf ("BaseMap destruction for 0x%09x [%d]\n", (unsigned long)this, msize) ;
#endif
    clear () ;
    delete [] htab ;
  }
  
	/////////////////////////////////
	//
	// Return a count of the number of bytes of memory
	// occupied by this Map.
	//
  int mem_used ()
    {
      return msize * (sizeof(slist<K,T>*) + sizeof(slist<K,T>))
	+ items * sizeof(slink<K,T>) ;
    }

protected:

  int abs (int x)    { return x < 0 ? -x : x ; }

  int hash (int ix)  { return abs(ix) % msize ; }

  int hash (char *s) { 
    int ix = 0 ;
    while (*s)
      ix += *s++ ;
    return abs(ix) % msize ;
  }
  
  int msize ;
  int items ;
  int th_low, th_high ;
  slist<K, T>* htab ;
};


/*--------------------------------------------------------------------+
 |                                                                    |
 |  template<class T> class Map_Iterator                              |
 |                                                                    |
 |    Used to iterate through all elements of a mapping, in pseudo    |
 |    random order.                                                   |
 |                                                                    |
 |    Example: to iterator through a symbol mapping called 'sym_map'  |
 |                                                                    |
 |      Map_Iterator<Symbol> next_sym (sym_map) ;                     |
 |      Symbol* s ;                                                   |
 |      while ( s = next_sym() ) { ... do whatever ... }              |
 |                                                                    |
 +--------------------------------------------------------------------*/

template <class K, class T> class Base_Map_Iterator {

friend class BaseMap<K, T> ;

public:

	/////////////////////////////////
	//
	//
	//
  Base_Map_Iterator () {
    msize = 0 ;
    h = 0 ;
    next = 0 ;
    bin = 0 ;
  }

	/////////////////////////////////
	//
	//
	//
  Base_Map_Iterator (BaseMap<K, T>& m) : next () {
    msize = m.msize ;
    h = m.htab ;
    for (bin = 0; bin < msize && h[bin].empty(); bin++) ;
    if (bin < msize) 
      next = h[bin] ;
  }

	/////////////////////////////////
	//
	//
	//
  T* operator() () {
    if (bin < msize)
    {
      T* v = next () ;
      if (v)
	return v ;
      else {
	while ( (++bin) < msize && h[bin].empty()) ;
	if (bin < msize) {
	  next = h[bin] ;
	  return next () ;
	}
	else
	  return 0 ;
      }
    }
    else
      return 0 ;
  }

private:
  int bin ;
  int msize ;
  slist<K, T>* h ;
  slist_iterator<K, T> next ;
};



template <class T> class Map : public BaseMap<int, T> {
public:
	/////////////////////////////////
	//
	//
	//
  Map (Map<T>& m) : BaseMap<int, T> (m) { }

  Map () : BaseMap<int, T> () { }
};

template <class T> class Map_Iterator : public Base_Map_Iterator<int, T> {
public:
	/////////////////////////////////
	//
	//
	//
  Map_Iterator (Map<T>& m) : Base_Map_Iterator<int, T> (m) { } ;
};

template <class T> class NameMap : public BaseMap<char*, T> {
public:
	/////////////////////////////////
	//
	//
	//
  NameMap (NameMap<T>& m) : BaseMap<char*, T> (m) { }

  NameMap () : BaseMap<char*, T> () { }
};

template <class T> class NameMap_Iterator : public Base_Map_Iterator<char*, T> {
public:
	/////////////////////////////////
	//
	//
	//
  NameMap_Iterator (NameMap<T>& m) : Base_Map_Iterator<char*, T> (m) { } ;
};



template <class T> class MapEntry {
public:
  T* value ;
  int id ;
  int index () { return id ; }
  int variant () { return 0 ; }
  MapEntry (int i, T* t) { id = i ; value = t ; }
};

/*--------------------------------------------------------------------+

CLASS
    ObjectMap

    Implements mappings from arbitrary class K objects to class T objects

DESCRIPTION
-----------------------------------------------------------------------*/

template <class K, class T> class ObjectMap : public Map<MapEntry<T> > {
public:
  
  ObjectMap () : Map<MapEntry<T> > () {}
  ObjectMap (ObjectMap<K,T> &m) : Map<MapEntry<T> > (m) {}
  
  T* operator[] (K &mapFrom) 
    {
      MapEntry<T> *entry = Map<MapEntry<T> >::operator[](mapFrom.index()) ;
      if (!entry)
	return 0 ;
      else
	return entry->value ;
    }
  
  T* operator[] (K *mapFrom) 
    {
      if (!mapFrom)
	return 0 ;
      MapEntry<T> *entry = Map<MapEntry<T> >::operator[](mapFrom->index()) ;
      if (!entry)
	return 0 ;
      else
	return entry->value ;
    }
  
  void add_entry (K &key, T &entry)
    {
      if (!(*this)[key])
	{
	  MapEntry<T> *newEntry = new MapEntry<T> (key.index(), &entry) ;
	  (*this) += newEntry ;
	}
    }

  T* retrieve (K &key)
    {
      T* entry = (*this)[key] ;
      if (!entry)
	{
	  entry = new T () ;
	  MapEntry<T> *newEntry = new MapEntry<T> (key.index(), entry) ;
	  (*this) += newEntry ;
	}
      return entry ;
    }

  T* retrieve (K *keyPtr)
    {
      if (!keyPtr)
	return 0 ;
      
      T* entry = (*this)[(*keyPtr)] ;
      if (!entry)
	{
	  entry = new T () ;
	  MapEntry<T> *newEntry = new MapEntry<T> (keyPtr->index(), entry) ;
	  (*this) += newEntry ;
	}
      return entry ;
    }
};

/*--------------------------------------------------------------------+

CLASS
    ObjectMap_Iterator

    Implements iterators over ObjectMaps from arbitrary class K objects to 
    class T objects.

DESCRIPTION
-----------------------------------------------------------------------*/

template <class K, class T> class ObjectMap_Iterator : public Map_Iterator<MapEntry<T> >
{
public:
  
  ObjectMap_Iterator (ObjectMap<K,T> &map) : Map_Iterator<MapEntry<T> > (map) {}
};

#endif // _map_h_
