/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: Symbols.cc,v 1.5 1999/07/29 15:46:17 npt Exp $

#include <ctype.h>
#include "Octave/Base/Symbols.h"
#include "Octave/Base/ParseContext.h"
#include "Octave/Fortran/Fortran.h"

/*--------------------------------------------------------------------+
 |                                                                    |
 |                      Symbol::set_fields                            |
 |                                                                    |
 | Sets context, tag, identifier, hash-table pointer and decl-type    |
 | fields for a new symbol. This cannot be declared in Symbols.h      |
 | as that creates a circular declaration dependence with the         |
 | hash_entry class (which references members of class Symbol)        |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::set_fields (PTR_CNTXT c, symbol_tag v, PTR_HASH h, int dtype = 0) { 
  pc = c ;
  symb_variant (v) ;
  ident = h->ident ;
  parent = h ;
  decl = dtype ;
  h->set_attr (this) ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                 Symbol::link_to_hashtable                          |
 |                                                                    |
 | Each symbol object may be contained within some scope S. The scope |
 | is represented as an object of class SymbolTable, which itself     |
 | contains scalar objects of class hash_entry. Each hash_entry       |
 | object contains a key (the identifier of the symbol) and a pointer |
 | to the corresponding Symbol object.                                |
 |                                                                    |
 | This routine establishes the link between a newly-created entry    |
 | in a SymbolTable (i.e. a pointer to a hash_entry) and the existing |
 | Symbol object. Any previous link that either object had, will be   |
 | erradicated by this function.                                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::link_to_hashtable (PTR_HASH h) {
  parent = h ;
  h->id_attr = this ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                     Symbol::append_id_list                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB Symbol::append_id_list (PTR_SYMB sym)
{
  PTR_SYMB t ;
  
  if (!sym)
    return this ;
  
  for (t = this; t->id_list; t = t->id_list) ;
  
  t->id_list = sym ;
  return this ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                        Symbol::process_type                        |
 |                                                                    |
 | This routine is called when the end of a TYPE ... ENDTYPE sequence |
 | has been parsed. The current parent points to the TYPE statement   |
 | and its bl_ptr1 list holds the field declaration statements.       |
 | The routine scans these statements, and in turn scans their lists  |
 | of field symbols, to link together the fields and to count the     |
 | number of fields.                                                  |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::process_type ()
{
  PTR_SYMB field_sym ;
  PTR_SYMB last_field ;
  PTR_EXPR sym_list ;
  PTR_STMT typedef_stmt ;
  PTR_STMT field_decl ;
  PTR_BLOCK field_list ;
  int count = 0 ;
  
  typedef_stmt = pc->cur_parent () ;
  field_list = typedef_stmt->then_part() ;
  
  last_field = field_list->hd()->node.fields.e1->node.list.item->node.fields.symbol;
  type->type_variant (T_STRUCT) ;
  
  /* The field derived class is a constraint put by the design. 
   * Maybe instead of T_STRUCT, T_RECORD might be better. 
   * But then it would be inconsistent with C++. */
  
  type->node.derived_class.first = last_field ;
  type->node.derived_class.original_class = typedef_stmt ;

  // Iterate through all statements in the first statement list
  // of the current control_parent.

  if (field_list)
  {
    StmtListIterator next_field (*field_list) ;
    
    while ( field_decl = next_field() )
    {
      sym_list = field_decl->node.fields.e1;
      while (sym_list != 0) 
      {
	count++ ;
	field_sym = sym_list->node.list.item->node.fields.symbol ;
	field_sym->node.fields.tag = FIELD_NAME ;
	last_field->node.fields.next = field_sym ;
	field_sym->node.fields.base_name = this ;
	sym_list = sym_list->node.list.next ;
      }
    }
  }
  type->node.derived_class.num_fields = count;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                         Symbol::component                          |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB Symbol::component (const char *fieldname)
{
  PTR_HASH entry ;
  PTR_SYMB t ;
  PTR_SCOPE class_scope ;
  
  class_scope = this->type->node.derived_class.original_class->scoping() ;
  
  entry = class_scope->find (fieldname, LOOK_LOCALLY, NO_MATCH, SEARCH_ONLY) ;
  
  if (entry && (t = entry->id_attr) &&
      (t->node.fields.tag == FIELD_NAME) && 
      (t->node.fields.base_name == this))
    return t ;
  else
    return 0 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                          Symbol::set_type                          |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::set_type (PTR_TYPE base, int ttag)
{
  PTR_SYMB temp = this ;
  PTR_SYMB last ;
  
  while (temp) {
    temp->symb_variant (VARIABLE_NAME) ;
    temp->type = base ;
    temp->node.var_decl.local = ttag ;
    last = temp ;
    temp = last->id_list ;
    last->id_list = 0 ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                        Symbol::install_const                       |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::install_const (PTR_EXPR e)
{
  PTR_SYMB temp = this, last ;
  
  while (temp && e) {
    temp->symb_variant (CONST_NAME) ;
    temp->type = e->type ;
    temp->node.const_value = e ;
    last = temp ;
    temp = last->id_list ;
    last->id_list = 0 ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     Symbol::install_param_list                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::install_param_list (PTR_SYMB ids, PTR_EXPR result, int func)
{
  int num_input = 0;
  PTR_SYMB temp = ids, last, result_sym_ptr;
  
  while (temp) {
    temp->variant (VARIABLE_NAME) ;
    temp->node.var_decl.next_in = temp->id_list ;
    temp->node.var_decl.local = IO ;
    ++num_input ;
    last = temp ;
    temp = last->id_list ;
    last->id_list = 0 ;
  }
  
  node.proc_decl.in_list = ids ;
  node.proc_decl.num_io = 0 ;
  node.proc_decl.num_input = num_input ;
  node.proc_decl.num_output = (func == FUNCTION_NAME) ;

  if (result) 
  {
    result_sym_ptr = result->node.fields.symbol ;
    node.fields.declared_name = result_sym_ptr ;
    result_sym_ptr->type = type ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                    Symbol::construct_entry_list                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_EXPR Symbol::construct_entry_list (PTR_SYMB symb_list, int kind)
{
  int num_input = 0 ;
  PTR_SYMB last_symb ;
  PTR_EXPR entry_param_list, l ;
  
  entry_param_list = 0 ;
  if (symb_list) 
  {
    l = pc->make_expr (VAR_REF, 0, 0, symb_list) ;
    entry_param_list = pc->set_ll_list (l, 0, EXPR_LIST) ;
    last_symb = symb_list ;
    symb_list = last_symb->id_list ;
    last_symb->id_list = 0 ;
    num_input++ ;
  }
  
  while (symb_list) 
  {
    l = pc->make_expr (VAR_REF, 0, 0, symb_list) ;
    entry_param_list = pc->set_ll_list (entry_param_list, l, EXPR_LIST) ;
    last_symb = symb_list ;
    symb_list = last_symb->id_list ;
    last_symb->id_list = 0 ;
    num_input++ ;
  }
  
  node.proc_decl.num_io = 0 ;
  node.proc_decl.num_input = num_input ;
  node.proc_decl.num_output = (kind == FUNCTION_NAME) ;
  return entry_param_list ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     Symbol::resolve_overloading                    |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB Symbol::resolve_overloading (PTR_EXPR argptr)
{
  PTR_SYMB current;
  
  if (variant() != INTERFACE_NAME)
    return this ;
  
  current = node.fields.declared_name ;
  while (current) 
  {
    current->reset_args () ;
    if (match_args (current, argptr))
      return current ;
    current = current->node.fields.declared_name ;
  }
  return 0 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                         Symbol::match_args                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

int Symbol::match_args (PTR_SYMB contender_proc, PTR_EXPR argptr)
{
  PTR_SYMB formal_argptr, formal_arg ;
  PTR_EXPR argref, actual_arg ;
  int match_status ;
  
  formal_argptr = contender_proc->node.fields.in_list ;
  while (argptr && formal_argptr)
  {
    argref = argptr->node.list.item ;
    if (argref->variant() == KEYWORD_ARG)
    {
      formal_arg = 
	contender_proc->find_keyword_arg (argref->node.fields.e1->node.string_val) ;
      actual_arg = argref->node.fields.e2 ;
    }
    else 
    {
      formal_arg = formal_argptr ;
      actual_arg = argref ;
    }
    match_status = match_an_arg (actual_arg, formal_arg) ;

    if (!match_status)
      return 0 ;
    
    formal_argptr = formal_argptr->node.var_decl.next_in ;
    argptr = argptr->node.list.next ;
  }
  
  return contender_proc->remainder_optional () ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                       Symbol::match_an_arg                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

int Symbol::match_an_arg (PTR_EXPR actual_arg, PTR_SYMB formal_arg)
{
  PTR_TYPE actual_arg_type, formal_arg_type ;
  
  actual_arg_type = actual_arg->type ;
  formal_arg_type = formal_arg->type ;
  
  if (actual_arg_type->variant() != formal_arg_type->variant()) 
    return 0 ;
  
  if ((actual_arg_type->variant() == T_DERIVED_TYPE) &&
      (actual_arg_type->name != formal_arg_type->name))
    return 0 ;
  
  if (actual_arg_type->variant() == T_ARRAY)
  {
    if (actual_arg_type->node.ar_decl.base_type->variant() !=
	formal_arg_type->node.ar_decl.base_type->variant())
      return 0 ;
    
    if ((actual_arg_type->node.ar_decl.base_type->variant() == T_DERIVED_TYPE)
	&& (actual_arg_type->node.ar_decl.base_type->name !=
	    formal_arg_type->node.ar_decl.base_type->name))
      return 0 ;
    
    if (actual_arg_type->node.ar_decl.num_dimensions !=
	formal_arg_type->node.ar_decl.num_dimensions)
      return 0 ;
  }
  
  formal_arg->decl = 1 ;
  return 1 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                         Symbol::reset_args                         |
 |                                                                    |
 +--------------------------------------------------------------------*/

void Symbol::reset_args ()
{
  PTR_SYMB temp;
  
  temp = node.fields.in_list ;
  while (temp != 0) 
  {
    temp->decl = 0 ;
    temp = temp->node.var_decl.next_in ;
  }
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                      Symbol::find_keyword_arg                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_SYMB Symbol::find_keyword_arg (string *keyword)
{
  PTR_SYMB temp;
  
  temp = node.fields.in_list;

  if (!keyword)
    return 0 ;
  else
    while (temp != 0) 
    {
      if (*keyword == temp->ident)
	return (temp);
      temp = temp->node.var_decl.next_in;
    }
  
  return 0 ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                     Symbol::remainder_optional                     |
 |                                                                    |
 +--------------------------------------------------------------------*/

int Symbol::remainder_optional ()
{
  PTR_SYMB temp = node.fields.in_list ;

  while (temp) 
  {
    if ((temp->decl == 0) && !(temp->attr & OPTIONAL_BIT))
      return 0 ;
    temp = temp->node.var_decl.next_in ;
  }
  
  return 1 ;
}     

/*--------------------------------------------------------------------+
 |                                                                    |
 |                       Symbol::subroutine_call                      |
 |                                                                    |
 +--------------------------------------------------------------------*/

PTR_STMT Symbol::subroutine_call (PTR_EXPR argptr)
{
  PTR_STMT stmt ;
  PTR_EXPR e ;
  PTR_SYMB current ;
  
  if (symb_variant() != INTERFACE_NAME) 
  {
    stmt = pc->make_stmt (PROC_STAT, this, argptr, 0, 0) ;
    return stmt ;
  }
  
  current = node.fields.declared_name ;
  while (current) 
  {
    current->reset_args () ;
    if (match_args (current, argptr))
    {
      e = pc->make_expr (PROC_CALL, argptr, 0, current);
      stmt = pc->make_stmt (OVERLOADED_PROC_STAT, this, argptr, e,0);
      return stmt ;
    }
    current = current->node.fields.declared_name ;
  }
  return 0 ;
}


/*--------------------------------------------------------------------+
 |                                                                    |
 |                   SymbolTable::set_implicit                        |
 |                                                                    |
 | Scans all symbols in a SymbolTable object, and if a symbol has its |
 | impl_type field set to IMPL, the type of that symbol is set        |
 | according to the types in the parameter array and the first char   |
 | of the symbol name.                                                |
 |                                                                    |
 +--------------------------------------------------------------------*/

void SymbolTable::set_implicit (PTR_TYPE impltype[26]) {
  NameMap_Iterator<hash_entry> next_symbol (*this) ;
  PTR_HASH h ;
  PTR_SYMB s ;

  while ( h = next_symbol () ) {
    s = h->id_attr ;
    PTR_TYPE newType = impltype[tolower(*(s->ident())) - 'a'] ;
    if (s->type == 0)
      s->type = newType ;
    else
      if ( s && (s->impl_type == IMPL) ) {
	switch (s->type->type_variant())
	  {
	  case T_ARRAY:
	    s->type->node.ar_decl.base_type = newType ;
	    break ;
	  case T_POINTER:
	    s->type->node.fields.base_type = newType ;
	    break ;
	  default:
	    s->type = newType ; 
	    break ;
	  }
      }
  }
}

void Symbol::CcUnparseParams (SourceListing& src, PTR_TYPE funType)
{
  SymbolList *sl = this ? this->node.cfun_decl.temps : 0 ;
  TypeList *tl   = funType? funType->node.fields.arg_types : 0 ;

  if (!this)
    {
      // The parameters are for an abstract function type, so only
      // unparse the parameter types (i.e. no declarator symbols).
      
      if (tl)
	{
	  TypeListIterator nextType (*tl) ;
	  PTR_TYPE t = nextType() ;
	  
	  while (t)
	    {
	      t->CcUnparseType (src) ;
	      t = nextType() ;
	      if (t)
		src << ", " ;
	    }
	}
    }
  else // actual function declarator... but param names may be missing
    if (sl)
      {
	SymbolListIterator nextSymb (*sl) ;
	PTR_SYMB s = nextSymb() ;
	TypeListIterator nextType (tl) ;
	PTR_TYPE st = nextType() ;
	
	while (s || st)
	  {
	    if (!s)
	      st->CcUnparseType (src) ;
	    else
	      if (!st)
		src << s->ident() ;
	      else
		{
		  PTR_TYPE dt = s->type->declspec () ;
		  if (dt && dt->type_variant() != T_UNKNOWN)
		    {
		      dt->CcUnparseType (src) ;
		      src << ' ' ;
		    }
		  s->CcUnparseDeclarator (src) ;
		}
	    s  = nextSymb() ;
	    st = nextType() ;
	    if (s || st)
	      src << ", " ;
	  }
      }
    else
      if (tl)
	{
	  TypeListIterator nextType (*tl) ;
	  PTR_TYPE t = nextType() ;
	  
	  while (t)
	    {
	      t->CcUnparseType (src) ;
	      t = nextType() ;
	      if (t)
		src << ", " ;
	    }
	}
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                 void CcUnparseDeclarator                            |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Symbol::CcUnparseDeclarator (SourceListing& src)
{
  if (!this)
    return ;

  // A declarator is a symbol with a defined type. This includes both
  // the declaration specifiers and the type modifiers (*, etc).
  // To print the declarator only, we just print the symbol in the
  // correct textual position with respect to its type modifiers.
  // The declaration specifiers are not printed.

  PTR_TYPE t = type ;
  TypeStack ts ;
  SymbolStack ss ;
  t->unp_absdecl (src, ts, ss, this) ;
  CcUnparseStack (src, ts, ss) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                  void CcUnparseFunDeclarator                        |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Symbol::CcUnparseFunDeclarator (SourceListing& src)
{
  if (!this)
    return ;

  TypeStack ts ;
  SymbolStack ss ;
  PTR_TYPE retn_type = type->node.fields.base_type->declspec() ;
  retn_type->CcUnparseType (src) ;
  src << ' ' ;
  type->unp_absdecl (src, ts, ss, this) ;
  CcUnparseStack (src, ts, ss) ;
}

/*--------------------------------------------------------------------+
 |                                                                    |
 |                   enumToText (symbol_tag)                          |
 |                                                                    |
 | Returns a const char* pointer to a text fields describing the      |
 | interpretation of the given symbol_tag.                            |
 |                                                                    |
 +--------------------------------------------------------------------*/

const char* enumToText (symbol_tag st)
{
  switch (st)
    {
    case DEFAULT_NAME:    return "DEFAULT_NAME" ;
    case CONST_NAME:      return "CONST_NAME" ;
    case ENUM_NAME:       return "ENUM_NAME" ;
    case FIELD_NAME:      return "FIELD_NAME" ;
    case VARIABLE_NAME:   return "VARIABLE_NAME" ;
    case TYPE_NAME:       return "TYPE_NAME" ;
    case PROGRAM_NAME:    return "PROGRAM_NAME" ;
    case PROCEDURE_NAME:  return "PROCEDURE_NAME" ;
    case VAR_FIELD:       return "VAR_FIELD" ;
    case LABEL_VAR:       return "LABEL_VAR" ;
    case FUNCTION_NAME:   return "FUNCTION_NAME" ;
    case MEMBER_FUNC:     return "MEMBER_FUNC" ;
    case CLASS_NAME:      return "CLASS_NAME" ;
    case UNION_NAME:      return "UNION_NAME" ;
    case STRUCT_NAME:     return "STRUCT_NAME" ;
    case LABEL_NAME:      return "LABEL_NAME" ;
    case COLLECTION_NAME: return "COLLECTION_NAME" ;
    case ROUTINE_NAME:    return "ROUTINE_NAME" ;
    case CONSTRUCT_NAME:  return "CONSTRUCT_NAME" ;
    case INTERFACE_NAME:  return "INTERFACE_NAME" ;
    case MODULE_NAME:     return "MODULE_NAME" ;
    case TEMPLATE_NAME:   return "TEMPLATE_NAME" ;
    case NAMESPACE_NAME:  return "NAMESPACE_NAME" ;
    case ANON_FIELD_NAME: return "ANON_FIELD_NAME" ;
    default: return "<undefined>" ;
    }
}
