/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _symbols_h_
#define _symbols_h_

#include "Strings.h"
#include "Sets.h"
#include "Map.h"
#include "Defs.h"
#include "Tags.h"

#define DEFAULT_HASH_SIZE 1931

enum Protection { PROT_PRIV = 0, PROT_PROT = 1, PROT_PUB = 2 } ;

class SymbolList ;
class SymbolListIterator ;
class SourceListing ;

class Symbol : public Tag {

friend class hash_entry ;
friend class Universe<Symbol> ;
friend class Datatype ;
friend class Expression ;
friend class Statement ;
friend class FortranContext ;

private:

  int		id;             // identifier for this symbol

public:

  PTR_CNTXT     pc;             // ptr back to containing context
  PTR_SYMB      thread;         // For Sage++/MARS compatibility only
  string	ident;          // symbol string
  PTR_HASH      parent;         // pointer to symbol table entry
  PTR_SYMB	id_list;	// used for making lists of ids
  PTR_TYPE	type;		// ptr to type for this symbol if declared
  PTR_TYPE      declType;       // ptr to type for symbol during declaration
  PTR_SCOPE	scope;		// scope level at which symbol is declared
  PTR_TRANS     trans ;         // pointer to transformation record
  int		attr;		// attributes of the symbol
  int		dovar;		// set if used as a loop control variable
  int		decl;		// field that the parser uses in keeping track
				// of declarations
  int	        impl_type;      // set to TYPED if symbol properly typed, IMPL
                                // otherwise
  union {
    PTR_EXPR  const_value;   // for all constant-valued symbols
    
    struct {  // for C and C++ enum-field and record field names
      int                 tag;
      int              offset;    
      PTR_SYMB  declared_name;  // used for friend construct
      PTR_SYMB           next;
      PTR_SYMB      base_name;  // name of record or enumerated type 
      PTR_EXPR restricted_bit;  // used by C and C++ for bit fields
    } field;
    
    struct {  // for variant fields
      int tag;
      int offset;
      PTR_SYMB next;
      PTR_SYMB base_name;
      PTR_EXPR variant_list;
    } variant_field;
    
    
    struct {  // for Fortran PROGRAM symbol
      PTR_SYMB       symb_list;
      PTR_FTN_LABEL label_list;
      PTR_STMT      prog_hedr;
    } prog_decl;
    
    struct {  // for Fortran SUBROUTINE symbols
      int	          seen;
      int	     num_input,
                    num_output, 
                        num_io;
      PTR_SYMB         in_list;
      PTR_SYMB        out_list;
      PTR_SYMB       symb_list;
      int	    local_size;
      PTR_FTN_LABEL label_list;
      PTR_STMT       proc_hedr;
      PTR_EXPR       call_list;
    } proc_decl;
    
    struct {  // for Fortran function symbols
      int	          seen;
      int	     num_input,
                    num_output, 
                        num_io;
      PTR_SYMB         in_list;
      PTR_SYMB        out_list;
      PTR_SYMB       symb_list;
      int	    local_size;
      PTR_FTN_LABEL label_list;
      PTR_STMT       func_hedr;
      PTR_EXPR       call_list;
    } func_decl;

    struct {               // for C and C++ FUNCTION_NAME symbols
      SymbolList *params;    // list of parameter symbols
      SymbolList  *temps;    // list of most-recently declared parameters
      int     stackspace;    // size of stackframe
      PTR_STMT func_hedr;    // ptr to header statement
    } cfun_decl ;

    struct {		   // for variable declaration
      int	   local;    // local or input or output or both param
      PTR_SYMB   next_in;    // for list of input parameters
      PTR_SYMB  next_out;    // for list of output parameters
      int         offset;    // stackframe offset
      int	   dovar;    // set if being used as DO control var
    } var_decl;
    
    struct {
      PTR_SCOPE    members;  // local scope for all class members
      PTR_EXPR   base_list;  // list of base classes
      Protection      prot;  // private, protected or public class
      SymbolList*  pubSyms;  // list of public members' symbols
      SymbolList* protSyms;  // list of protected members' symbols
      SymbolList* privSyms;  // list of private members' symbols
    } class_info ;

    struct {    // for C++ member function symbols
      int	        seen;
      int	   num_input, 
                  num_output, 
                      num_io;
      PTR_SYMB       in_list;
      PTR_SYMB      out_list;
      PTR_SYMB     symb_list;
      int         local_size;
      PTR_STMT     func_hedr;
      PTR_EXPR     call_list;
      int	         tag;
      int	      offset;
      PTR_SYMB declared_name;
      PTR_SYMB          next;
      PTR_SYMB     base_name;
    } member_func	    ;
    
    struct {    // a common unification of symbol fields
      int	          seen;
      int	     num_input,
                    num_output,
                        num_io;
      PTR_SYMB         in_list;
      PTR_SYMB        out_list;
      PTR_SYMB       symb_list;
      int	    local_size;
      PTR_FTN_LABEL label_list;
      PTR_STMT       func_hedr;
      PTR_EXPR       call_list;
      int	           tag;
      int	        offset;
      PTR_SYMB   declared_name;
      PTR_SYMB            next;
      PTR_SYMB       base_name;
    } fields ;
    
  } node ;
  
public:
  
  /////////////////////////////
  //
  // Default constructor has no actions.
  //
  Symbol () : Tag () { }

  /////////////////////////////
  //
  // Member creation constructor always requires an ID
  //
  Symbol (int i) : Tag (0) 
  { 
    id = i ;
    parent = 0 ;
    id_list = thread = 0 ;
    type = 0 ;
    scope = 0 ;
    trans = 0 ;
    dovar = decl = attr = impl_type = 0 ;
    node.fields.seen = 0 ;
    node.fields.num_input = 0 ;
    node.fields.num_output = 0 ;
    node.fields.num_io = 0 ;
    node.fields.in_list = 0 ;
    node.fields.out_list = 0 ;
    node.fields.symb_list = 0 ;
    node.fields.local_size = 0 ;
    node.fields.label_list = 0 ;
    node.fields.func_hedr = 0 ;
    node.fields.call_list = 0 ;
    node.fields.tag = 0 ;
    node.fields.offset = 0 ;
    node.fields.declared_name = 0 ;
    node.fields.next = 0 ;
    node.fields.base_name = 0 ;
  }

  void set_fields (Symbol& s)
  {
    pc = s.context() ;
    symb_variant (s.symb_variant()) ;
    ident = new string (s.ident) ;
    type = s.type ;
    attr = s.attr ;
    dovar = s.dovar ;
    decl = s.decl ;
    impl_type = s.impl_type ;
    // This is currently an imperfect copy of a Symbol...
  }

  void set_fields (PTR_CNTXT, symbol_tag, PTR_HASH, int) ;

  void set_fields (PTR_CNTXT c, symbol_tag v, const char *s) {
    pc = c ; 
    symb_variant (v) ;
    string x (s) ;
    ident = x ;
  }

  void set_fields (PTR_CNTXT c, symbol_tag v, string *s) {
    pc = c ;
    symb_variant (v) ;
    ident = s ;
  }

  void link_to_hashtable (PTR_HASH h) ;

  void operator= (PTR_SYMB source) {
     this->attr = source->attr;
     this->node.fields.seen = source->node.fields.seen;
     this->node.fields.num_input = source->node.fields.num_input;
     this->node.fields.num_output = source->node.fields.num_output;
     this->node.fields.num_io = source->node.fields.num_io;
     this->node.fields.in_list = source->node.fields.in_list;
     this->node.fields.out_list = source->node.fields.out_list;
     this->node.fields.local_size = source->node.fields.local_size;
     this->node.fields.label_list = source->node.fields.label_list;
     this->node.fields.func_hedr = source->node.fields.func_hedr;
     this->node.fields.call_list = source->node.fields.call_list;
     this->node.fields.tag = source->node.fields.tag;
     this->node.fields.offset = source->node.fields.offset;
     this->node.fields.declared_name = source->node.fields.declared_name;
     this->node.fields.next = source->node.fields.next;
     this->node.fields.base_name = source->node.fields.base_name;
  }

  int index () { return id ; }
  // int variant() is provided by the base class Tag

  PTR_CNTXT context () { return pc ; }

  PTR_SYMB append_id_list (PTR_SYMB) ;
  PTR_SYMB component (const char *) ;
  PTR_SYMB resolve_overloading (PTR_EXPR) ;
  PTR_SYMB find_keyword_arg (string *) ;

  PTR_STMT subroutine_call (PTR_EXPR) ;

  PTR_EXPR construct_entry_list (PTR_SYMB, int) ;

  void process_type () ;
  void startproc (int) ;
  void install_const (PTR_EXPR) ;
  void install_param_list (PTR_SYMB, PTR_EXPR, int) ;
  void reset_args () ;
  void set_type (PTR_TYPE, int) ;

  int match_args (PTR_SYMB, PTR_EXPR) ;
  int match_an_arg (PTR_EXPR, PTR_SYMB) ;
  int remainder_optional () ;

    ////////////////////////////////////
    //
    // Annotates this Symbol with a pointer to a transformation
    // record describing the transformation applied to this node.
    //
  void note_transform (PTR_TRANS t) { trans = t ; }

    ////////////////////////////////////
    //
    // Returns a pointer to a TransHist object describing
    // describing the transformation applied to this node.
    // May return a null pointer if there was no transformation.
    //
  PTR_TRANS transformation ()       { return trans ; }

  // GROUP: unparsing functions

    ////////////////////////////////////
    //
  void CcUnparseParams (SourceListing&, PTR_TYPE) ;

    ////////////////////////////////////
    //
  void CcUnparseDeclarator (SourceListing&) ;

    ////////////////////////////////////
    //
  void CcUnparseFunDeclarator (SourceListing&) ;
};


class hash_entry {

friend class Symbol ;
friend class SymbolTable ;
friend class ScopeStack ;
friend class FortranParser ;

private:

  string    ident ;		// identifier of this entry
  PTR_SYMB  id_attr ;		// symbolic information for this entry
  PTR_SCOPE decl_scope ;	// ptr to scope at which it is declared
  char      is_alias ;          // entry is an alias if this is non-null.

public:

  char *index ()               { return ident() ; }
  int variant ()               { return id_attr->variant() ; }

  PTR_SYMB symbol ()           { return id_attr ; }
  const char* identifier ()    { return ident() ; }
  string *name ()              { return &ident ;   }
  PTR_SCOPE scope ()	       { return decl_scope ; }

  void set_attr (PTR_SYMB sym) { id_attr = sym ; }
  void set_attr (PTR_SCOPE sc) { decl_scope = sc ; }
  void set_attr (PTR_SYMB sym, PTR_SCOPE sc) { id_attr = sym ; decl_scope = sc ;}
  void set_attr (int ia) { is_alias = ia ; }

  PTR_TYPE lookup_type (PTR_CNTXT) ;

  hash_entry (const char *s, PTR_SYMB sym = 0, PTR_SCOPE sc = 0, int ia = 0) : ident (s) 
  { 
    id_attr = sym ;
    decl_scope = sc ;
    is_alias = ia ;
  }

  hash_entry (string& s, PTR_SYMB sym = 0, PTR_SCOPE sc = 0, int ia = 0) : ident (s)
  { 
    id_attr = sym ;
    decl_scope = sc ;
    is_alias = ia ;
  }

  hash_entry (string* s, PTR_SYMB sym = 0, PTR_SCOPE sc = 0, int ia = 0) : ident (s)
  { 
    id_attr = sym ;
    decl_scope = sc ;
    is_alias = 0 ;
  }
} ;


class SymbolList : public slist<int,Symbol> { } ;

class SymbolListIterator : public slist_iterator<int,Symbol>
{
public:
  SymbolListIterator (SymbolList &sl) : slist_iterator<int,Symbol> (sl) {}
};


class SymbolTable : public NameMap<hash_entry> {

friend class FortranParser ;

public:
  
  void set_implicit (PTR_TYPE[26]) ;

  PTR_HASH lookup (char *name, int varnt) { 
    if (this)
      if (varnt)
	return lookup_variant (name, varnt) ;
      else
	return this->operator[] (name) ;
    else
      return 0 ;
  }

  SymbolTable () { }

  SymbolTable (NameMap<hash_entry>& st) : NameMap<hash_entry> (st) { }

  ~SymbolTable () { }

  void DeleteEntries () { zap () ; }

} ;

class SymbolTable_Iterator : public NameMap_Iterator<hash_entry> {

public:
  
  SymbolTable_Iterator (NameMap<hash_entry>& stab) : NameMap_Iterator<hash_entry> (stab) {} 
} ;

class SymbolStack : public stack<Symbol> { };

#endif // _symbols_h_
